import time
import random
import os
import requests
import traceback
import codecs
import json
import color_console as cons
from time import sleep
from selenium import webdriver
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.common import exceptions as selenium_exceptions


# 浏览器驱动器错误异常，需要关闭后重新创建
class DriverErrorException(Exception):
    pass

# 元素定位异常
class LocationErrorException(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)

# 关闭所有弹窗，没有则忽略
def try_close_all_alert(driver, logger):
    try:
        result = EC.alert_is_present()(driver)
        if result:
            result.accept()
    except:
        logger.log_warn('Fail to find and accept alert box: ' + str(traceback.format_exc()))


# 日志记录类
class Logger:
    def __init__(self, log_file, result_file):
        self.log_file = log_file
        self.result_file = result_file
        self.level = 0
        self.default_colors = cons.get_text_attr()
        self.default_bg = self.default_colors & 0x0070
        self.default_fg = self.default_colors & 0x0007

    def to_previous_level(self):
        if self.level > 0:
            self.level -= 1        

    def to_next_level(self):
        self.level += 1

    def clear_level(self):
        self.level = 0

    def write(self, message, is_warn, is_error, is_include_log, is_include_result):
        indent = ' ' * (self.level * 3)
        prefix = ''

        if is_warn:
            prefix = '!!! '
            cons.set_text_attr(cons.FOREGROUND_YELLOW | cons.BACKGROUND_BLACK | cons.FOREGROUND_INTENSITY | cons.BACKGROUND_INTENSITY)
        elif is_error:
            cons.set_text_attr(cons.FOREGROUND_RED | cons.BACKGROUND_BLACK | cons.FOREGROUND_INTENSITY | cons.BACKGROUND_INTENSITY)
            prefix = '### '
        else:
            cons.set_text_attr(self.default_colors)

        message = indent + prefix + message
        print(message)
        if is_include_log:
            self.log_file.write(message)
            self.log_file.write('\r\n')
        if is_include_result:
            self.result_file.write(message)
            self.result_file.write('\r\n')

    def console(self, message):
        self.write(message, False, False, False, False)

    def log(self, message):
        self.write(message, False, False, True, False)

    def log_warn(self, message):
        self.write(message, True, False, True, False)

    def log_error(self, message):
        self.write(message, False, True, True, False)

    def result(self, message):
        self.write(message, False, False, True, True)

    def result_error(self, message):
        self.write(message, False, True, True, True)

# 执行刷新
def refresh(driver, target, logger):

    target_url = target[1]
    locations = target[2]

    # 确保关闭了提示框
    try:
        driver.get("about:blank")
    except selenium_exceptions.UnexpectedAlertPresentException:
        driver.switch_to_alert().accept()
        driver.get("about:blank")
    except:
        logger.log_warn('Page load blank error: ' + str(traceback.format_exc()))
        raise DriverErrorException()

    sleep(0.5)
    try:
        driver.get(target_url)
    except selenium_exceptions.TimeoutException:
        logger.log_warn('Page load timeout: ' + str(traceback.format_exc()))
        try:
            driver.execute_script('window.stop()')
        except:
            logger.log_warn('Fail to execute script window.stop(): ' + str(traceback.format_exc()))

    search_window = driver.current_window_handle

    if not locations:
        # 无点击，仅刷新时，在页面多停留一段时间
        sleep(7)
        return True

    first_location = locations[0]
    try:
        WebDriverWait(driver, 10, 0.2).until(EC.presence_of_element_located(first_location))
    except:
        logger.log_warn('Fail to wait the first locating ' + str(first_location) + ': ' + str(traceback.format_exc()))

    element = driver
    for location in locations:
        logger.log('Locate to: ' + str(location))

        try:
            if location[0] == "id":
                element = element.find_element_by_id(location[1])
            elif location[0] == "css selector":
                element = element.find_element_by_css_selector(location[1])
            else:
                raise LocationErrorException("Not support locate: " + str(location))

            if element.tag_name.lower() == "iframe":
                driver.switch_to_frame(element)
                element = driver
        except:
            raise LocationErrorException("Locate " + str(location) + " error: " + str(traceback.format_exc()))

    logger.console("Find target: " + str(element))
    
    try:
        element.click()
    except selenium_exceptions.WebDriverException:
        # 异常信息可能为：Message: unknown error: Element is not clickable at point (x, y)
        # 尝试滚动之后再点击
        driver.execute_script("arguments[0].scrollIntoView(false);", element)
        sleep(2)
        element.click()

    # 延迟7秒，为点击后执行脚本留点时间
    sleep(7)

    try:
        # 检测提示框，并关闭
        try_close_all_alert(driver, logger)

        # 关闭除搜索窗口之外的窗口
        for handle in driver.window_handles:
            if handle != search_window:
                driver.switch_to_window(handle)
                driver.close()
    except:
        logger.log_warn('Fail to close other windows: ' + str(traceback.format_exc()))
    finally:
        # 切换回主窗口，并尝试关闭主窗口的提示框
        try:
            driver.switch_to_window(search_window)
        except:
            logger.log_warn('Fail to switch to search window: ' + str(traceback.format_exc()))
        try_close_all_alert(driver, logger)

    return True


# 格式化时长
def format_time(seconds):
    seconds = round(seconds)
    if seconds < 60:
        return str(seconds) + 's'
    elif seconds >= 60 and seconds < 3600:
        return str(int(seconds/60)) + 'm' + format_time(seconds%60)
    else:
        return str(int(seconds/3600)) + 'h' + format_time(seconds%3600)


# 递归删除文件夹
def remove_dir_and_files(target_dir):
    target_dir = target_dir.replace('\\', '/')

    # 删除目录
    if(os.path.isdir(target_dir)):
        # 先删除子目录
        for subdir in os.listdir(target_dir):
            remove_dir_and_files(os.path.join(target_dir, subdir))
        # 然后删除目录本身
        if(os.path.exists(target_dir)):
            os.rmdir(target_dir)
    # 删除文件
    elif(os.path.exists(target_dir)):
        os.remove(target_dir)


# 创建浏览器驱动
def create_driver(downloads_dir):
    user_agents = [
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.62 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/56.0.2924.87 Safari/537.36",
        "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0)",
        "Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.1; WOW64; Trident/7.0; SLCC2; .NET CLR 2.0.50727; .NET CLR 3.5.30729; .NET CLR 3.0.30729; .NET4.0C; .NET4.0E)",
        "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36 OPR/48.0.2685.50",
        "Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko",
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36",
        "Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; WOW64; Trident/6.0)"
    ]

    user_agent = user_agents[int(random.random() * 1000000)%len(user_agents)]
    chrome_options = webdriver.ChromeOptions()
    chrome_options.add_argument("headless")
    chrome_options.add_argument("disable-gpu")
    chrome_options.add_argument('window-size=1280,800')
    chrome_options.add_argument('user-agent=' + user_agent)
    chrome_options.add_argument("--disable-extensions")
    chrome_options.add_argument("--no-sandbox")
    chrome_options.add_argument('--log-level=3')
    chrome_options.add_experimental_option('prefs', {'profile.default_content_settings.popups': 0, 'download.default_directory': downloads_dir})
    driver = webdriver.Chrome(chrome_options = chrome_options)
    driver.set_page_load_timeout(30)
    driver.set_script_timeout(30)
    driver.maximize_window()
    return driver


# 主程序
def main():
    start_time = time.time()
    file_name = time.strftime("%Y%m%d_%H%M%S", time.localtime()) + '.txt'
    
    # 日志记录文件
    log_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'logs')
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    log_file_path = os.path.join(log_dir, file_name)
    log_file =  codecs.open(log_file_path, "w", "utf-8")

    # 结果记录文件
    result_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'results')
    if not os.path.exists(result_dir):
        os.makedirs(result_dir)
    result_file_path = os.path.join(result_dir, file_name)
    result_file =  codecs.open(result_file_path, "w", "utf-8")

    logger = Logger(log_file, result_file)

    targets = []
    max_retry_count_get_refresh_list = 10
    for index in range(max_retry_count_get_refresh_list):
        try:
            targets_json_bytes = requests.get("http://download.guanli360.com/quanguan/seo/refreshes.txt").content
            targets_json_string = targets_json_bytes.decode('utf-8-sig')
            targets = json.loads(targets_json_string)
            break
        except:
            retry_count = index + 1
            logger.log_warn("Fail to download refresh list (" + str(retry_count) + " times): " + str(traceback.format_exc()))
            # 已到重试上限
            if retry_count == max_retry_count_get_refresh_list:
                logger.result_error("Fail to download refresh list.")
                break
            # 延迟30秒后继续执行
            sleep(30)

    random.shuffle(targets)

    '''
    targets = [
                  ["PC6下载", "http://www.pc6.com/softview/SoftView_240872.html", [["id", "showding"]]]
                   ,["华军ERP", "http://www.onlinedown.net/soft/577314.htm", [["css selector", ".dincai > iframe"], ["css selector", "#dincai > div.dincaibox.dingbox > span.icon.din"]]]
                   #,["华军CRM", "http://www.onlinedown.net/soft/577315.htm", [["css selector", "#dincai > iframe"], ["css selector", "#dincai > div.dincaibox.dingbox > span.icon.din"]]]
                   #,["华军OA", "http://www.onlinedown.net/soft/577316.htm", [["css selector", "#dincai > iframe"], ["css selector", "#dincai > div.dincaibox.dingbox > span.icon.din"]]]

              ]
    '''

    downloads_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'downloads')
    if not os.path.exists(downloads_dir):
        os.makedirs(downloads_dir)

    driver = create_driver(downloads_dir)

    logger.result('===============================================================================')
    logger.result('===                             开始执行刷新任务                            ===')
    logger.result('===============================================================================')
    logger.result('开始时间：' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + "，任务数：" + str(len(targets)))
    logger.result('-------------------------------------------------------------------------------')
    
    success_count = 0
    for target in targets:
        target_title = target[0]
        target_url = target[1]
        logger.result('>> ' + target_title + ': ' + target_url)
        try:
            logger.to_next_level()
            is_ok = False
            try:
                refresh(driver, target, logger)
                is_ok = True
            except DriverErrorException:
                logger.log_warn('Recreate driver')
                driver.quit()
                driver = create_driver(downloads_dir)
                refresh(driver, target, logger)
                is_ok = True
            finally:
                if is_ok:
                    logger.result("$$$ Refresh success!")
                else:
                    logger.result_error("Fail to refresh.")
            success_count += 1
        except:
            logger.log_error('Fail to execute refreshing: ' + str(traceback.format_exc()))
        finally:
            logger.clear_level()

    try:
        driver.quit()
        remove_dir_and_files(downloads_dir)
    except:
        logger.log_warn("Fail to delete downloads dir (" + downloads_dir + "): " + str(traceback.format_exc()))

    end_time = time.time()

    timeconsuming_one = 0 if not targets else (end_time - start_time)/len(targets)
    logger.result('-' * 80)
    logger.result(' 共' + str(len(targets)) + '个刷新任务，成功' + str(success_count) + '个，耗时' + format_time(end_time - start_time) + '，平均耗时' + format_time(timeconsuming_one))
    
    log_file.close()
    result_file.close()

    os.popen('notepad "' + result_file_path + '"')

# 运行
main()
